import argparse


import sortasurvey
from sortasurvey import pipeline
from sortasurvey import INPDIR, OUTDIR


def main():
    parser = argparse.ArgumentParser(
                                     description="sort-a-survey: automated, optimizable and reproducible target selection",
                                     prog="sort-a-survey",
    )
    parser.add_argument('-version', '--version',
                        action='version',
                        version="%(prog)s {}".format(sortasurvey.__version__),
                        help="Print version number and exit."
    )

    # In the parent parser, we define arguments and options common to all subcommands
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('-info', '--info', '-survey', '--survey',
                               dest='path_survey',
                               help='Path to csv containing survey information',
                               type=str,
                               default='%s/survey_info.csv'%INPDIR,
    )
    parent_parser.add_argument('-in', '--in', '-inpdir', '--inpdir', '-input', '--input',
                               dest='inpdir',
                               help='Path to survey information files and input',
                               default=INPDIR,
                               type=str,
    )
    parent_parser.add_argument('-no', '--no', '-nono', '--nono', '-ignore', '--ignore',
                               dest='path_ignore',
                               help='Path to save results to',
                               default='%s/no_no.csv'%INPDIR,
                               type=str,
    )
    parent_parser.add_argument('-out', '--out', '-outdir', '--outdir', '-output', '--output',
                               dest='outdir',
                               help='Path to save results to',
                               default=OUTDIR,
                               type=str,
    )
    parent_parser.add_argument('-prio', '--prio', '-priority', '--priority',
                               dest='path_priority',
                               help='Path to main planet sample to select from',
                               default='%s/high_priority.csv'%INPDIR, 
                               type=str,
    )
    parent_parser.add_argument('-sample', '--sample','-samples', '--samples',
                               dest='path_sample',
                               help='Path to main planet sample to select from',
                               default='%s/TOIs_perfect.csv'%INPDIR, 
                               type=str,
    )
    parent_parser.add_argument('-verbose', '--verbose',
                               dest='verbose',
                               help='Turn off verbose output (default=False)',
                               default=True, 
                               action='store_false',
    )

    sub_parser = parser.add_subparsers(title='subcommands', dest='subcommand')

    # Setting up
    parser_setup = sub_parser.add_parser('setup', parents=[parent_parser],
                                         description='Easy setup for directories and files')
    parser_setup.set_defaults(func=pipeline.setup)

    # Running pySYD in regular mode

    parser_run = sub_parser.add_parser('rank', help='Rank targets for the survey', 
                                       parents=[parent_parser])
    parser_run.add_argument('-iter', '--iter', '-step', '--step', '-iterations', '--iterations', '-steps', '--steps', 
                            dest='iter', 
                            help='Number of selection process iterations (default=1)',
                            default=1, 
                            type=int,
    )
    parser_run.add_argument('-prog', '--prog', '-progress', '--progress',
                            dest='progress',
                            help='Turn off progress bar (default=True). Only activates for > 1 iteration',
                            default=True, 
                            action='store_false',
    )
    parser_run.add_argument('-hpn', '--hpn', '-hours', '--hours', '-hourspernight', '--hourspernight',
                            dest='hours',
                            type=float,
                            default=10.,
                            help="Number of hours per night. (default=10)",
    )
    parser_run.add_argument('-night', '--night', '-pool', '--pool',  '-nights', '--nights',
                            dest='nights',
                            type=float,
                            default=50.,
                            help="Total number of allocated nights for survey.",
    )
    parser_run.add_argument('-save', '--save', 
                            dest='save',
                            help='Turn off the saving of output data products and figures (default=True)',
                            default=True, 
                            action='store_false',
    )
    parser_run.add_argument('-show', '--show',
                            dest='show',
                            help='Show output figures (default=False)',
                            default=False, 
                            action='store_true',
    )
    parser_run.add_argument('-sp', '--sp', '-special', '--special',
                            dest='special',
                            help='For TKS, this is for a more complicated, 2D selection process',
                            nargs='*',
                            type=str,
                            default=None,
    )

    parser_run.set_defaults(func=pipeline.rank)

    args = parser.parse_args()
    args.func(args)



if __name__ == '__main__':

    main()


import numpy as np
import pandas as pd
from tqdm import tqdm
pd.set_option('mode.chained_assignment', None)


from sortasurvey import observing


class Survey:
    """
    Loads in survey information and the vetted survey sample.

    Parameters
    ----------
    args : argparse.Namespace
        command line arguments

    Attributes
    ----------
    verbose : bool
        verbose output
    iter : int
        number of selection process iterations. Default is `1` (via args.iter).
    step : int
        iteration number
    programs : pandas.DataFrame
        pandas dataframe containing survey information
    sciences : pandas.DataFrame
        copy of the 'programs' pandas dataframe
    sample : pandas.DataFrame
        pandas dataframe containing the sample to select targets from
    candidates : pandas.DataFrame
        copy of the 'sample' pandas dataframe
    track : dict
        logs each iteration of the target selection

    """
    
    def __init__(self, args, path_sample='info/TOIs_perfect.csv', path_survey='info/survey_info.csv',
                 path_priority='info/high_priority.csv', path_ignore='info/no_no.csv', iter=1,
                 hours_per_night=10., pool=50., progress=True):

        self.verbose = args.verbose
        self.save = args.save
        self.outdir = args.outdir
        self.iter = args.iter
        self.progress = args.progress
        self.track = {}
        for n in np.arange(1,self.iter+1):
            self.track[n] = {}
        if self.verbose:
            print('\n -- prioritization starting --\n\n   - loading sample and survey science information')
        self.get_sample(args)
        self.get_programs(args)
        self.get_seeds()
        if self.iter > 1:
            self.emcee = True
            if self.verbose and self.progress:
                self.pbar = tqdm(total=self.iter)
        else:
            self.emcee = False


    def get_sample(self, args, dec=-30., ruwe=2.):
        """
        Fixes the sample based on specific survey needs.
        For TKS as a whole, this only required that the target
        is observable (dec > -30.) and possessed a reasonable
        Gaia RUWE metric (ruwe < 2.).

        """
        df = pd.read_csv(args.path_survey, comment="#")
        self.programs = df.programs.values.tolist()
        sample = pd.read_csv(args.path_sample)
        self.path_sample = args.path_sample
        self.sample = sample.query("dec > %f and ruwe < %f"%(dec, ruwe))
        self.remove_bad()
        self.add_columns()
        # science-case-specific metrics
        self.get_sc3_info()
        self.get_counts()
        self.candidates = self.sample.copy()
        

    def remove_bad(self, disp=['FP','EB','NEB','BEB','SV','BD','NPC','SB1','SB2','FA']):
        """
        Removes unfavorable targets from the survey. In this case, this includes false 
        alarms and/or false positives, including nearby/blended eclipsing binaries as well
        as spectroscopic false positives (e.g., SB1, SB2).

        """
        for bad in disp:
            self.sample.query("disp != '%s'"%bad, inplace=True)
        self.sample.query("drop == False", inplace=True)
        self.sample.query("finish == False", inplace=True)
    

    def add_columns(self, cols=["npl","select_DG","in_other_programs","n_select"]):
        """
        Adds in additional columns that might be relevant for the target selection.

        """

        cols += ["in_%s"%program for program in self.programs]
        for col in cols:
            if col == 'npl':
                self.sample[col] = self.sample.groupby('tic')['tic'].transform('count')
            else:
                self.sample[col] = [0]*len(self.sample)


    def get_counts(self):
        """
        Compute the number of targets that passed the different vetting steps.

        """
        query = self.sample.drop_duplicates(subset='tic')
        new_query = query.query("photo_vetting == 'passed' and spec_vetting != 'failed' and spec_vetting != 'do not observe' and ao_vetting != 'failed'")
        self.passed_tks = len(query)
        self.passed_vet = len(new_query)
        if self.verbose:
            print('   - %d targets make the standard survey cuts'%self.passed_tks)
            print('   - %d have also passed various vetting steps'%self.passed_vet)
            print('   - ranking algorithm initialized')


    def get_programs(self, args):
        """
        Get the survey programs from the provided csv (via args.path_survey)

        """

        programs = pd.read_csv(args.path_survey, comment="#")
        programs.set_index('programs', inplace=True, drop=False)
        nono = pd.read_csv(args.path_ignore)
        priority = pd.read_csv(args.path_priority)

        for program in programs.index.values.tolist():
            # get initial allocation
            programs.loc[program,'total_time'] = (programs.loc[program,'allocations']/(np.sum(programs.allocations)))*args.nights*args.hours
            if not np.isnan(programs.loc[program,'remaining_hours']):
                programs.loc[program,'total_time'] += programs.loc[program,'remaining_hours']
            programs.loc[program,'remaining_hours'] = programs.loc[program,'total_time']
            # adjust filter for priority/ignore targets
            high_priority = [float(target) for target in priority[program].values if target != '-']
            no_no = [float(target) for target in nono[program].values if target != '-']
            for toi in (high_priority + no_no):
                programs.loc[program,'filter'] += " and toi != %.2f"%toi
            if programs.loc[program,'n_maximum'] != -1:
                programs.loc[program,'n_targets_left'] = programs.loc[program,'n_maximum']
            else:
                query = self.sample.query(programs.loc[program,'filter'])
                targets = query.toi.values.tolist() + high_priority
                targets = [int(np.floor(each)) for each in targets]
                programs.loc[program,'n_targets_left'] = len(list(set(targets)))
            # get the prioritization metric
            prioritize_by=[]
            for each in programs.loc[program,'prioritize_by'].strip().split('|'):
                prioritize_by.append(each)
            programs.at[program,'prioritize_by'] = prioritize_by
            ascending_by=[]
            for each in programs.loc[program,'ascending_by'].strip().split('|'):
                if each == 'True' or str(each) == 'True' or str(each) == 'TRUE':
                    ascending_by.append(True)
                else:
                    ascending_by.append(False)
            programs.at[program,'ascending_by'] = ascending_by
        programs['pick_number'] = np.zeros(len(programs)).astype('int64')
        programs['n_targets_left'] = programs['n_targets_left'].astype('int64')
        programs.rename(columns={'programs':'name'}, inplace=True)
        programs.drop(columns=['allocations'], inplace=True)
        programs = programs.to_dict('index')
        for program in programs:
            programs[program]['high_priority'] = [float(target) for target in priority[program].values if target != '-']
        programs = pd.DataFrame.from_dict(programs, orient='index', columns=['name','method','filter','prioritize_by','ascending_by','remaining_hours','n_maximum','total_time','high_priority','n_targets_left','pick_number'])
        programs.to_csv('%s_copy.%s'%(args.path_survey.split('.')[0],args.path_survey.split('.')[-1]))
        self.programs = programs.copy()
        self.sciences = self.programs.copy()
        if args.special is not None:
            self.get_special(args)
        else:
            self.special = {}


    def get_special(self, args, special=[]):
        """
        Get any special science programs that require a more complicated selection process.
        For TKS, SC1B uses a 2D selection process and needs to update the filter at each selection
        step.

        """
        for program in args.special:
            if program.startswith('o') or program.startswith('O'):
                special.append('T%s'%program.upper())
            else:
                special.append('SC%s'%program.upper())
        sp={}
        for program in special:
            sp[program]={}
        self.special = sp.copy() 


    # science-case-specific functions
    def get_sc3_info(self, include_qlp=True, mask=None):
        """
        Get any special science programs that require a more complicated selection process.
        For TKS, SC1B uses a 2D selection process and needs to update the filter at each selection
        step.

        """
        if not include_qlp:
            mask = self.sample['source'] == 'spoc'
        self.calculate_TSM(mask=mask)
        sc3_df = self.sample.copy()
        sc3_mask = pd.notnull(sc3_df['TSM'])
        rp_bins = 10**(np.linspace(0,1,6))
        rp_bins[-1] = 11.2
        sinc_bins = 10**(np.linspace(-1,4,6))
        teff_bins = np.array([2500,3900,5200,6500])
        bins = [rp_bins, sinc_bins, teff_bins]
        sc3_df['rt_5sig'] = sc3_df['rt_5sig'].replace(0., 1e-2)
        sc3_mask &= pd.notnull(sc3_df['rt_5sig'])
        sc3_df = sc3_df[sc3_mask]
        sc3_df['X'] = sc3_df['TSM']/sc3_df['rt_5sig']
        sc3_df = self.sc3_binning_function(sc3_df, bins, sort_val='X')
        sc3_df['SC3_bin_rank'] = sc3_df['SC3_bin_rank'].replace(0., np.nan)
        cols_to_use = ['toi','SC3_bin_rank']
        self.sample = self.sample.merge(sc3_df[cols_to_use], how='left', on='toi')


    def calculate_TSM(self, mask=None):
        """
        Get any special science programs that require a more complicated selection process.
        For TKS, SC1B uses a 2D selection process and needs to update the filter at each selection
        step.

        """
        self.sample['TSM'] = [np.nan]*len(self.sample)
        if mask is None:
            mask = np.ones(len(self.sample), dtype=bool)
        for key in ['rp','mp','a_to_R','teff','jmag','r_s']:
            mask &= pd.notnull(self.sample[key])
        def rp_to_scale_factor(rp):
            if rp < 1.5:
                scale_factor = 0.19
            elif rp > 1.5 and rp < 2.75:
                scale_factor = 1.26
            elif rp > 2.75 and rp < 4:
                scale_factor = 1.28
            else:
                scale_factor = 1.15
            return scale_factor
        scale_factors = self.sample.loc[mask,'rp'].apply(rp_to_scale_factor)
        Teqs = self.sample.loc[mask,'teff']*np.sqrt(np.reciprocal(self.sample.loc[mask,'a_to_R'])*np.sqrt(0.25))
        numerator = scale_factors*np.power(self.sample.loc[mask, 'rp'], 3)*Teqs*np.power(10, -1*self.sample.loc[mask,'jmag']/5)
        denominator = self.sample.loc[mask,'mp']*np.square(self.sample.loc[mask,'r_s'])
        self.sample.loc[mask,'TSM'] = numerator/denominator


    def sc3_binning_function(self, df, bins, sort_val='TSM', num_to_rank=5):
        """
        Get any special science programs that require a more complicated selection process.
        For TKS, SC1B uses a 2D selection process and needs to update the filter at each selection
        step.

        """
        pre_bin = df.assign(
            rp_bin = pd.cut(df['rp'],bins=bins[0],labels = [1,2,3,4,5]),
            sinc_bin = pd.cut(df['sinc'],bins=bins[1],labels = [1,2,3,4,5]),
            teff_bin = pd.cut(df['teff'],bins=bins[2],labels = [1,2,3])
        )
        binned_df = pre_bin.dropna(subset=['rp_bin','sinc_bin','teff_bin']).\
                groupby(['rp_bin','sinc_bin','teff_bin']).apply(lambda _pre_bin:\
                _pre_bin.sort_values(by=[sort_val],ascending=False))\
                .reset_index(level = 3,drop=True)
        all_idx = binned_df.index.tolist()
        unique_idx = []
        for element in all_idx:
            if element not in unique_idx:
                unique_idx.append(element)
        binned_df['SC3_bin_rank'] = np.zeros(len(binned_df))
        for idx in unique_idx:
            bin_items = len(binned_df.loc[idx].sort_values(sort_val,ascending=False).iloc[0:num_to_rank]['toi'])
            for i in range(1, num_to_rank+1):
                if bin_items == i and bin_items <= num_to_rank:
                    for j in range(i):
                        binned_df.loc[binned_df['toi'] == binned_df.loc[idx].sort_values\
                                (sort_val,ascending=False).iloc[0:num_to_rank]['toi'].iloc[j],'SC3_bin_rank'] = j+1
                elif bin_items > num_to_rank:
                    for j in range(num_to_rank):
                        binned_df.loc[binned_df['toi'] == binned_df.loc[idx].sort_values\
                                (sort_val,ascending=False).iloc[0:num_to_rank]['toi'].iloc[j],'SC3_bin_rank'] = j+1
        return binned_df


    def init_2D_filters(self, program):
        """
        Sets up filters for a more complicated selection criteria.
        This is relevant for SC1B, exploring the possible connection 
        between stellar flux and gaseous envelopes. To study this
        effect, the goal is to randomly but uniformly sample in
        place radius vs. incident flux space. This is done by setting
        up equal bin sizes in linear(rp)-log(sinc) parameter space 
        for a total of nbins=12.

        """
        count=0
        choices={}
        if program == 'SC1B':
            y_bins = np.arange(1,5)
            x_bins = np.logspace(0,2,5)
            science_filter = "rp > %f and rp < %f and period > %f and period < %f"
            n_bins = (len(y_bins)-1)*(len(x_bins)-1)
            for i in range(n_bins):
                x = i%(len(x_bins)-1)
                y = i%(len(y_bins)-1)
                filter = science_filter%(y_bins[y],y_bins[y+1],x_bins[x],x_bins[x+1])
                query = self.candidates.query(filter)
                query = query.drop_duplicates(subset='tic')
                if not query.empty:
                    choices[count] = {}
                    choices[count]['targets_left'] = len(query)
                    choices[count]['filter'] = filter
                    count += 1
            choices = pd.DataFrame.from_dict(choices, orient='index')
            if count != n_bins:
                choose = np.arange(1,count+1,1)
            else:
                choose = np.arange(1,n_bins+1,1)
            # it will randomize the bin filters for one entire round (using all 12 bins)
            choose = np.random.permutation(choose)[:]
            choices = choices.set_index(choose)
            choices = choices.sort_index()
        return choices


    def get_2D_filter(self, program):
        """
        This procedure is called for science cases that wish to uniformly 
        and randomly sample over a 2-dimensional space. 

        """
        pick_number = self.special[program]['pick_number']
        choices = self.special[program]['choices']
        while True:
            n_bins = len(choices)
            # first check if the bin has any targets left
            if pick_number%n_bins == 0:
                # special case for the final bin in the round of choices
                check = choices.loc[n_bins,'targets_left']
                if check != 0:
                    # if it does, use the filter
                    filter = choices.loc[n_bins,'filter']
                    # reset filter choices
                    self.reset_2D_filters(program)
                    break
                else:
                    # if not, reset filter choices
                    self.reset_2D_filters(program)
            else:
                check = choices.loc[pick_number%n_bins,'targets_left']
                if check != 0:
                    # select filter
                    filter = choices.loc[pick_number%n_bins,'filter']
                    break
            # keep going until an appropriate bin/filter is selected
            pick_number += 1
        pick_number += 1
        self.special[program]['pick_number'] = pick_number
        self.sciences.loc[program,'filter'] = filter


    def reset_2D_filters(self, program):
        """
        When all bins have been used, they need to be reset (which includes)
        randomizing the bin order selection.

        """
        choices = self.special[program]['choices']
        new_df = choices.query('targets_left != 0')
        n_bins = len(new_df)
        choose = np.arange(1,n_bins+1,1)
        choose = np.random.permutation(choose)[:]
        choices = new_df.set_index(choose)
        self.special[program]['choices'] = choices.copy()


    def check_2D_overlap(self, pick, program):
        """
        Multi-planet systems (i.e. the same observed target) can fall into more than
        one bin so this needs to be checked at each step, as ewll as the number of 
        available targets per bin.

        """
        choices = self.special[program]['choices']
        tic = pick.tic
        for i in choices.index.values.tolist():
            query = self.candidates.query(choices.loc[i,'filter'])
            if tic in query.tic.values.tolist():
                choices.loc[i, 'targets_left'] -= 1
        self.sciences.loc[program, 'n_targets_left'] = np.sum(choices.targets_left.values.tolist())+1
        self.special[program]['choices'] = choices.copy()
        self.special[program]['stuck'] = 0


    def update_targets(self):
        start = np.array([0]*len(self.candidates))
        for science in self.sciences.index.values.tolist():
            start += self.candidates['in_%s'%science].values.tolist()
        self.candidates['in_other_programs'] = start
 

    def reset_track(self):
        """
        Get any special science programs that require a more complicated selection process.
        For TKS, SC1B uses a 2D selection process and needs to update the filter at each selection
        step.

        """
        self.reset_programs()
        self.track[self.n][0] = {}
        for program, hours in zip(self.sciences.index.values.tolist(), self.sciences.remaining_hours.values.tolist()):
            self.track[self.n][0][program] = round(hours,3)
        self.track[self.n][0]['total_time'] = round(np.sum(self.sciences.remaining_hours.values.tolist()),3)
        self.priority = 1
        self.i = 1
        np.random.seed(self.seeds[self.n-1])
        if self.verbose and self.progress and self.iter > 1 and self.n != 1:
            self.pbar.update(1)


    def reset_programs(self):
        """
        Get any special science programs that require a more complicated selection process.
        For TKS, SC1B uses a 2D selection process and needs to update the filter at each selection
        step.

        """
        # make copies of the original dataframes, thus resetting the information
        self.candidates = self.sample.copy()
        self.sciences = self.programs.copy()
        if self.special != {}:
            for program in self.special:
                self.special[program]['stuck'] = 0
                self.special[program]['pick_number'] = 1
                choices = self.init_2D_filters(program)
                self.special[program]['choices'] = choices


    def add_program_pick(self, pick, program):
        """
        Get any special science programs that require a more complicated selection process.
        For TKS, SC1B uses a 2D selection process and needs to update the filter at each selection
        step.

        """
        self.track[self.n][self.i] = {}
        self.track[self.n][self.i]['program'] = program
        self.track[self.n][self.i]['program_pick'] = self.sciences.loc[program,'pick_number']+1
        self.track[self.n][self.i]['toi'] = float(pick.toi)
        self.track[self.n][self.i]['tic'] = int(pick.tic)


    def update_program_hours(self):
        """
        Get any special science programs that require a more complicated selection process.
        For TKS, SC1B uses a 2D selection process and needs to update the filter at each selection
        step.

        """
        for program, hours in zip(self.sciences.index.values.tolist(), self.sciences.remaining_hours.values.tolist()):
            self.track[self.n][self.i][program] = round(hours,3)
        self.track[self.n][self.i]['total_time'] = round(np.sum(self.sciences.remaining_hours.values.tolist()),3)


    def update(self, pick, program):
        """
        Update appropriate information and tables.

        """
        if program in self.special:
            self.check_2D_overlap(pick, program)
        self.add_program_pick(pick, program)
        self.sciences.loc[program,'n_targets_left'] -= 1
        self.sciences.loc[program,'pick_number'] += 1
        self = observing.check_observing(self, pick, program)
        if not int(pick.in_other_programs):
            net = {program:-1.*(float(pick.actual_cost)/3600.)}
            self.track[self.n][self.i]['overall_priority'] = self.priority
            self.candidates.loc[self.candidates['tic'] == int(pick.tic),'priority'] = int(self.priority)
            self.priority += 1
        else:
            net = observing.adjust_costs(self, pick, program)
            idx = self.candidates.loc[self.candidates['tic'] == int(pick.tic)].index.values.tolist()[0]
            self.track[self.n][self.i]['overall_priority'] = int(self.candidates.loc[idx,'priority'])
        for key in net.keys():
            self.sciences.loc[key,'remaining_hours'] += net[key]
        self.update_program_hours()
        self.candidates.loc[self.candidates['tic'] == int(pick.tic),'in_%s'%program] = 1
        self.update_targets()
        self.i += 1


    def get_seeds(self):
        """
        Ensures reproducibility due to the instrinsic randomness of the algorithm.

        """
        self.seeds = [2222, 5531, 5348, 9632, 3755, 3401, 1061, 9307, 2033, 2114, 3103, 8120, 5442, 9179, 3165, 6114, 8757, 8574, 8078, 7724, 9056, 9066, 8423, 5278, 663, 4542, 6448, 7261, 6999, 7212, 3832, 3199, 6444, 1704, 8872, 2743, 9163, 1293, 8458, 5782, 7144, 9339, 3961, 9127, 4105, 3209, 7662, 5592, 4672, 2365, 8214, 3725, 2088, 1234, 6984, 2756, 3962, 7279, 9686, 112, 8936, 8807, 4149, 2535, 1541, 1422, 7991, 6445, 4384, 570, 9719, 5834, 5372, 1376, 1192, 1499, 8653, 730, 5469, 7541, 6546, 4002, 5677, 9251, 5459, 630, 908, 9074, 2675, 9517, 1015, 5272, 6846, 6820, 4516, 5632, 5671, 2126, 4440, 9670, 7768, 1405, 5330, 1854, 3156, 6949, 1119, 5257, 2999, 4251, 9674, 5362, 5009, 7526, 8293, 4518, 8641, 1365, 2492, 5061, 4804, 2710, 8823, 6637, 9382, 7928, 9219, 7840, 895, 5647, 3966, 6452, 9027, 8673, 1006, 469, 5056, 42, 8067, 7571, 3304, 6795, 9131, 6327, 5781, 5336, 4484, 5137, 3231, 4465, 91, 5135, 3303, 1890, 7593, 359, 6051, 1236, 9967, 3149, 9913, 3114, 9267, 3049, 6089, 6439, 828, 8893, 7708, 6766, 2818, 8745, 8791, 3639, 461, 3917, 8917, 2863, 1865, 9410, 1851, 617, 7563, 915, 1773, 4997, 6121, 8540, 6358, 1630, 5468, 8585, 4959, 8115, 6337, 355, 1977, 4800, 6831, 932, 1028, 8232, 1381, 3260, 2937, 7031, 6310, 5348, 2172, 3321, 4422, 1195, 2021, 481, 731, 5566, 9719, 7468, 9499, 1326, 4071, 7660, 6583, 5067, 5693, 2933, 8679, 9988, 550, 2599, 5536, 3081, 4429, 3592, 8140, 1398, 1481, 6823, 9006, 9264, 6037, 95, 9807, 2768, 4792, 7417, 6095, 8049, 79, 5070, 1457, 3099, 736, 2332, 2228, 146, 3862, 2153, 7800, 8664, 625, 2393, 88, 780, 4266, 9412, 4973, 426, 7742, 4593, 408, 7296, 1981, 867, 7636, 2455, 3519, 3093, 882, 7396, 815, 7717, 4792, 3103, 2747, 290, 8302, 2124, 2516, 3170, 8224, 3693, 5721, 3599, 9778, 5903, 8544, 69, 7648, 4860, 212, 517, 3765, 1401, 8722, 1689, 3281, 3061, 9293, 4954, 4584, 3357, 6380, 5266, 8972, 5578, 9289, 859, 486, 3746, 7928, 7240, 2861, 7615, 651, 5633, 4687, 7439, 2572, 1999, 1476, 5806, 1966, 9249, 3439, 4559, 6899, 5633, 1973, 6469, 1636, 4922, 5059, 7772, 3907, 7410, 1822, 9659, 8230, 3643, 9106, 9524, 8971, 2887, 705, 4252, 6198, 1420, 9063, 5272, 9641, 195, 5217, 1819, 2286, 431, 5379, 26, 7690, 7241, 3735, 2987, 1490, 2807, 5059, 6556, 5921, 3949, 6128, 606, 7636, 1451, 4598, 2446, 9877, 635, 876, 9594, 1742, 5887, 5355, 365, 8197, 7919, 6969, 9736, 1703, 8703, 3358, 8321, 6817, 3617, 9069, 6406, 3938, 3077, 6166, 1546, 4393, 1026, 9479, 2568, 1787, 1434, 8390, 3844, 4028, 5643, 9291, 5072, 8022, 7260, 1209, 5579, 6860, 2871, 2662, 4769, 7361, 7427, 8737, 1608, 6613, 7941, 5619, 6949, 3217, 4204, 1439, 3439, 4521, 4761, 4089, 2066, 9623, 3076, 9230, 1503, 9896, 7110, 2152, 1291, 1339, 5088, 2959, 8092, 5381, 7283, 8831, 8448, 6775, 5414, 5871, 2728, 8828, 6320, 3294, 7953, 4157, 5654, 6890, 5134, 45, 6881, 4237, 9561, 913, 9990, 9667, 650, 1353, 2963, 3896, 4368, 8162, 5630, 5889, 9093, 5298, 17, 7958, 6417, 7574, 6461, 7446, 8398, 5486, 7742, 7503, 1740, 6987, 2238, 1159, 6552, 7968, 440, 1671, 7755, 9214, 1099, 7801, 4910, 878, 3278, 667, 1813, 7540, 2082, 3182, 5580, 3256, 9619, 5890, 8902, 9635, 2516, 864, 823, 9222, 6156, 5011, 7191, 4584, 4112, 9991, 110, 2361, 2709, 6469, 9592, 9668, 6788, 7505, 4174, 3119, 5693, 429, 6224, 3174, 6134, 6902, 9692, 2620, 1532, 7973, 5644, 6105, 2495, 1368, 9342, 3747, 9358, 1039, 311, 5382, 7309, 2482, 1889, 1162, 5620, 8439, 5487, 975, 4845, 4641, 7027, 747, 1016, 5728, 6175, 5252, 598, 4920, 5544, 6273, 9336, 8096, 8059, 2467, 1098, 72, 372, 737, 4500, 2736, 7458, 3742, 3156, 8420, 5311, 8532, 7186, 9113, 7041, 6658, 2370, 2733, 8258, 139, 6127, 4489, 692, 5627, 8139, 9744, 9773, 3674, 9103, 9896, 897, 2939, 8342, 3031, 4991, 3110, 3845, 2214, 5184, 7482, 3367, 5030, 9570, 7613, 1394, 1491, 2570, 5573, 9688, 2731, 8333, 6764, 4922, 4886, 8623, 2301, 8688, 9286, 832, 439, 2502, 8934, 5356, 6584, 6322, 6958, 1542, 9526, 9040, 10000, 8659, 1672, 610, 4050, 6616, 7105, 6073, 9004, 5102, 7781, 1615, 8225, 2511, 3862, 6110, 9382, 5402, 1501, 1972, 6596, 2496, 2523, 2710, 3515, 4024, 7273, 6509, 1913, 8888, 5892, 6173, 1836, 7008, 1328, 6628, 6840, 126, 3190, 4511, 5644, 8944, 6386, 8863, 5022, 5361, 3799, 6701, 750, 785, 2069, 6609, 6429, 7252, 5477, 2309, 9163, 7957, 4056, 8866, 6815, 8583, 2891, 6979, 1242, 3795, 4564, 1785, 1292, 3009, 8132, 3837, 5357, 5549, 9030, 9177, 7603, 3764, 347, 3695, 3836, 7269, 1196, 5401, 4362, 4053, 9416, 2994, 6420, 8527, 7178, 1084, 1582, 967, 7636, 3565, 6510, 4259, 6769, 7106, 1102, 2072, 5721, 4149, 1459, 4861, 39, 1404, 44, 7296, 3745, 2023, 3162, 4885, 9147, 2716, 4395, 9489, 9240, 9882, 3761, 2755, 1862, 9856, 404, 7118, 8258, 5581, 1477, 4694, 463, 598, 9566, 9119, 6289, 5209, 6703, 4719, 8622, 9687, 8361, 5639, 812, 6559, 9332, 6663, 5722, 3930, 8141, 6207, 7787, 1572, 6012, 6052, 609, 6106, 606, 3013, 3915, 6504, 7301, 5596, 1644, 4915, 5623, 943, 1779, 1028, 5734, 8674, 6440, 5126, 5988, 4179, 2955, 9198, 4068, 7912, 6211, 8559, 7260, 193, 7662, 8317, 7231, 181, 1482, 9115, 9971, 4519, 4073, 4300, 2938, 5456, 2939, 3906, 6385, 5011, 9510, 1227, 8649, 4715, 6021, 5418, 3568, 3571, 622, 5414, 1137, 5375, 6263, 4930, 2352, 8565, 6277, 563, 1113, 4847, 1355, 3215, 5297, 8883, 8371, 6917, 5952, 3448, 6154, 52, 2204, 763, 919, 7355, 2095, 8809, 2362, 7590, 5591, 4950, 2817, 7882, 9214, 7549, 5118, 1046, 2298, 7458, 8277, 339, 2443, 8941, 2072, 324, 1350, 5502, 2501, 2680, 5925, 9935, 6294, 5578, 6686, 5888, 5921, 2690, 8177, 2405, 5438, 6754, 2331, 5550, 1591, 9183, 3714, 1097, 7171, 1552, 117, 1135, 1067, 4952, 6742, 3960, 1489, 4201, 2390, 7777, 979, 2114, 9652, 7569, 5795, 4386, 7838, 6684, 1262, 7700, 8091, 1979, 3566, 6058, 5834, 5443, 50, 1658, 6047, 4273, 9477, 3761, 5953, 2732, 7142, 81, 4457, 4703, 5029, 9303, 1936, 468, 5888, 7253, 6611, 1883, 8285, 7267, 2028, 2842, 9302, 4262, 812, 2696, 9879, 2992, 4865, 5031, 6292, 2439, 2261, 8345, 9926, 8960, 2960, 4199]


import numpy as np
import pandas as pd
from scipy.optimize import brentq
pd.set_option('mode.chained_assignment', None)



def cost_function(df, method, include_archival=True, overhead=2.*60, lower=3.*60, upper=20.*60):
    """
    Estimates the total amount of time needed on sky for a given target, which 
    is highly dependent on the instrument using to collect the data
    PLEASE NOTE: this might need to be adjusted for other surveys

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe containing a target's vmag and other required information in the cost function
    method : str
        observing method of a particular program
    
    Returns
    -------
    rem_time : float
        the remaining time (in seconds) needed to achieve your specified science.

    """

    counts = method.split('=')[-1]
    nobs = int(float((method.split('-')[1]).split('=')[-1]))
    if counts == 'ramp':
        counts = exp_ramp(df['vmag'], c1=125., c2=60.)
    else:
        counts = float(counts)
    # estimate how much time a target would take, given a program's observing method
    exp = exposure_time(df['vmag'], counts, iodine=True)
    # make a cut at a survey's maximum allowable exposure time per observation
    exp = np.clip(exp, lower, upper)
    # include archival data in total time estimates
    if include_archival:
        rem_nobs = int(nobs-df['nobs'])
        if rem_nobs < 0:
            rem_nobs = 0
            rem_time = 0.
        else:
            rem_time = exp*rem_nobs+overhead*rem_nobs
        # if a template has not been acquired for a target yet
        if not df['template']:
            exp = exposure_time(df['vmag'], 250., iodine=False)
            exp = np.clip(exp, 3.*60, 45.*60)
            rem_time += (exp+overhead)
    else:
        # if not including archival data, consider the time to acquire a template
        rem_time = exp*nobs+overhead*nobs
        exp = exposure_time(df['vmag'], 250., iodine=False)
        exp = np.clip(exp, 3.*60, 45.*60)
        rem_time += (exp+overhead)
    return rem_time


def exposure_time(vmag, counts, iodine=False, t1=110., v1=8., exp1=250., iodine_factor=0.7):
    """
    Expected exposure time based on the scaling from a canonical exposure time
    of 110s to get to 250k on 8th mag star with the iodine cell in the light
    path
    
    Parameters
    ----------
    expcounts : float 
        desired number of counts 
        250 = 250k, 10 = 10k (CKS) i.e. SNR = 45 per pixel.
    iodine (bool) : is iodine cell in or out? If out, throughput is higher by 30%
    
    Returns
    -------
    exptime : float
        exposure time [seconds]
    
    """

    # flux star / flux 8th mag star
    fluxfactor = 10.0**(-0.4*(vmag-v1)) 
    exptime = t1/fluxfactor 
    exptime *= counts/exp1
    if iodine == False:
        exptime *= iodine_factor
    return exptime


def exposure_counts(vmag, exptime, **kwargs):
    """
    Inverse of `exposure_time.` Given a magnitude and an exposure
    time, how many counts will be collected?

    Parameters
    ----------
    vmag : float
        Johnson V mag
    exptime : float
        exposure time in seconds
        **kwargs : keyword arguments passed to exposure_time
    
    Returns
    -------
    counts : float
        expected number of photon counts (x1000)

    """

    f = lambda expcounts : exposure_time(vmag, expcounts, **kwargs) - exptime
    _counts = brentq(f,0,200000,)

    return _counts


def exp_ramp(vmag, v1=10.5, v2=12.0, c1=250., c2=60.):
    """
    Calculates exposure counts based on a minimum (v1) and maximum (v2)
    magnitude limits, with a linear ramp between the two magnitude limits.

    Parameters
    ----------
    vmag : float 
        target magnitude
    v1 : float 
        below this mag targets get full counts (c1)
    v2 : float
        fainter than this mag targets get c2
    c1 : float
        expcounts (k) for bright targets
    c2 : float
        expcounts (k) for the faint limit

    Returns
    -------
    counts : float
        expected number of photon counts (x1000)

    """

    if vmag <= v1:
        return c1

    if vmag >= v2:
        return c2

    exp_level = np.interp(vmag, xp=[v1, v2], fp=[np.log10(c1), np.log10(c2)])
    counts = 10.**exp_level

    return counts

   

def get_actual_costs(program, programs, query):
    """
    Called during each sampling step to recompute the most up-to-date costs
    for a given target based on past algorithm selections

    Parameters
    ----------
    program : str
        selected program that is making the selection
    programs : pandas.DataFrame
        survey program dataframe
    query : pandas.DataFrame
        all targets from the vetted sample relevant for the selected program
    
    Returns
    -------
    query : pandas.DataFrame
        the relevant vetted sample updated with actual target costs

    """
    actual_costs=[]
    for index in query.index.values.tolist():
        df = query.loc[index]
        costs = []
        for science in programs.index.values.tolist():
            if query.loc[index,'in_%s'%science]:
                time = cost_function(df, programs.loc[science,'method'])
                costs.append(time)
        time = cost_function(df, programs.loc[program,'method'])
        costs.append(time)
        if float(np.sum(costs)) != 0.:
            fraction = costs[-1]/np.sum(costs)
            actual_costs.append(fraction*max(costs))
        else:
            actual_costs.append(0.)
    query['actual_cost'] = np.array(actual_costs)
    return query


def adjust_costs(survey, pick, program):
    """
    Based on the selected program's pick, calculates the credit/debit amounts
    for any program that has selected the same target.

    Parameters
    ----------
    survey : survey.Survey
        class object containing both the vetted sample (via survey.candidates) and the
        survey programs (via survey.sciences)
    pick : pandas.DataFrame
        a single row dataframe containing information on the selected program's current pick
    program : str
        the selected program
    cases : List[str]
        other programs that have selected a given target in past iterations
    costs : List[float]
        the cost of a target given a specific program's observing strategy
    
    Returns
    -------
    net : Dict[str,float]
        a python dictionary whose keys are other programs to credit or debit back, pointing
        to the amount of time to credit or debit the program back

    """
    cases=[]
    costs=[]
    index = survey.candidates.index[survey.candidates['tic'] == int(pick.tic)].tolist()
    df_pick = survey.candidates.loc[index[0]]
    for science in survey.sciences.index.values.tolist():
        if pick['in_%s'%science]:
            time = cost_function(df_pick, survey.sciences.loc[science]['method'])
            cases.append(science)
            costs.append(time)
    cases.append(program)
    if float(np.sum(costs)) == 0.:
        net_costs = -1.*np.zeros(len(cases))
    else:
        frac = costs/np.sum(costs)
        old_costs = list((max(costs)/3600.)*frac)
        old_costs.append(0)
        time = cost_function(df_pick, survey.sciences.loc[program]['method'])
        costs.append(time)
        new_frac = costs/np.sum(costs)
        new_costs = np.array((max(costs)/3600.)*new_frac)
        net_costs = -1.*(new_costs - np.array(old_costs))
    net = dict(zip(cases,net_costs))
    return net


def check_observing(survey, pick, program):
    """
    Based on programs that have selected a given target, this will figure out which
    observing method requires the most observing time and then scales other program's
    costs based on that.

    Parameters
    ----------
    survey : survey.Survey
        class object containing both the vetted sample (via survey.candidates) and the
        survey programs (via survey.sciences)
    pick : pandas.DataFrame
        a single row dataframe containing information on the selected program's current pick
    program: str
        the selected program
    
    Returns
    -------
    survey : survey.Survey
        class object with updated nobs_goal information, if applicable
    """
    method = survey.sciences.loc[program, "method"]
    nobs_goal = int(float((method.split('-')[1]).split('=')[-1]))
    idx = survey.candidates.loc[survey.candidates['tic'] == int(pick.tic)].index.values.tolist()
    if nobs_goal > survey.candidates.loc[idx[0], 'nobs_goal']:
        for index in idx:
            survey.candidates.loc[index, 'nobs_goal'] = nobs_goal
    return survey


import os
import datetime
import numpy as np
import pandas as pd
pd.set_option('mode.chained_assignment', None)


from sortasurvey import observing


def pick_program(programs):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    cdf = list(np.cumsum(programs.remaining_hours.values.tolist())/np.sum(programs.remaining_hours.values.tolist()))
    cdf.insert(0, 0.)
    pick = np.random.random()
    for i in range(len(cdf)-1):
        if cdf[i] < pick and cdf[i+1] > pick:
            break
    program = programs.index.values.tolist()[i]
    return program


def make_data_products(survey):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    if survey.verbose and not survey.emcee:
        query = survey.df.query('in_other_programs != 0')
        query = query.drop_duplicates(subset = 'tic')
        print("   - algorithm took %d seconds to run"%(int(survey.ranking_time)))
        print('   - %d targets were selected'%len(query))
    if survey.save:
        if not survey.emcee:
            survey = make_directory(survey)
        else:
            if survey.n == 1:
                survey = make_directory(survey)
            os.makedirs('%s/%d/'%(survey.path_save,survey.n))
    survey = make_final(survey)
    survey = make_ranking_steps(survey)
    survey = assign_priorities(survey)
#    survey = final_costs(survey)
    survey = program_overlap(survey)
    get_stats(survey)


def make_directory(survey, i=1):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    if survey.verbose and not survey.emcee:
        print('   - making data products')
    now = datetime.datetime.now()
    name = now.strftime("%m-%d-%y")
    newdir = '%s/%s-%d'%(survey.outdir,name,i)
    if not os.path.exists(newdir):
        os.makedirs(newdir)
    else:
        while os.path.exists(newdir):
            i += 1
            newdir = '%s/%s-%d'%(survey.outdir,name,i)
        os.makedirs(newdir)
    survey.path_save = newdir
    return survey
          
  
def make_final(survey):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    changes = survey.df.query('in_SC2A == 1 and in_other_programs == 1')
    method = survey.programs.loc['SC2A', 'method']
    for i in changes.index.values.tolist():
        df_temp = changes.loc[i]
        nobs_goal = int(float((method.split('-')[1]).split('=')[-1]))
        survey.df.loc[i, "nobs_goal"] = nobs_goal
        tottime = observing.cost_function(df_temp, method, include_archival=False)
        survey.df.loc[i, "tot_time"] = round(tottime/3600.,3)
        remaining_nobs = int(nobs_goal - survey.df.loc[i, "nobs"])
        if remaining_nobs < 0:
            remaining_nobs = 0
        survey.df.loc[i, "rem_nobs"] = remaining_nobs
        lefttime = observing.cost_function(df_temp, method)
        survey.df.loc[i, "rem_time"] = round(lefttime/3600.,3)
    changes = survey.df.query('in_SC4 == 1 and in_other_programs == 1')
    method = survey.programs.loc['SC4', 'method']
    for i in changes.index.values.tolist():
        df_temp = changes.loc[i]
        nobs_goal = int(float((method.split('-')[1]).split('=')[-1]))
        survey.df.loc[i, "nobs_goal"] = nobs_goal
        tottime = observing.cost_function(df_temp, method, include_archival=False)
        survey.df.loc[i, "tot_time"] = round(tottime/3600.,3)
        remaining_nobs = int(nobs_goal - survey.df.loc[i, "nobs"])
        if remaining_nobs < 0:
            remaining_nobs = 0
        survey.df.loc[i, "rem_nobs"] = remaining_nobs
        lefttime = observing.cost_function(df_temp, method)
        survey.df.loc[i, "rem_time"] = round(lefttime/3600.,3)
    if survey.save:
        if survey.emcee:
            survey.df.to_csv('%s/%d/TOIs_perfect_final.csv'%(survey.path_save, survey.n), index=False)
        else:
            survey.df.to_csv('%s/TOIs_perfect_final.csv'%survey.path_save, index=False)
        if survey.verbose and not survey.emcee:
            print('   - copy of updated TOI spreadsheet saved to %s'%survey.path_save)
    survey.final = survey.df.copy()
    return survey
      
  
def make_ranking_steps(survey):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    reorder = get_columns('track', survey)
    track = pd.DataFrame.from_dict(survey.track[survey.n], orient='index')
    df = pd.DataFrame(columns = reorder)
    for column in reorder:
        df[column] = track[column]
    if survey.save:
        if survey.emcee:
            df.to_csv('%s/%d/ranking_steps.csv'%(survey.path_save,survey.n))
        else:
            df.to_csv('%s/ranking_steps.csv'%survey.path_save)
        if survey.verbose and not survey.emcee:
            print('   - ranking steps of the algorithm have been saved to %s/ranking_steps.csv'%survey.path_save)
    survey.ranking_steps = df.copy()
    return survey
    
    
def assign_priorities(survey, obs={}, m=1):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    track_sorted = survey.ranking_steps.sort_values(by = ['overall_priority'])
    for q in list(set(track_sorted.overall_priority.values.tolist())):
        if not np.isnan(q):
            prior = track_sorted[track_sorted["overall_priority"] == q]
            obs[m] = {}
            obs[m]['tic'] = int(prior.tic.values.tolist()[0])
            obs[m]['toi'] = int(np.floor(prior.toi.values.tolist()[0]))
            obs[m]['programs'] = prior.program.values.tolist()
            m += 1
    observed = pd.DataFrame.from_dict(obs, orient='index')
    observed.reset_index(inplace = True)
    observed = observed.rename(columns = {'index':'overall_priority'})
    if survey.save:
        if survey.emcee:
            observed.to_csv('%s/%d/observing_priorities.csv'%(survey.path_save,survey.n))
        else:
            observed.to_csv('%s/observing_priorities.csv'%survey.path_save, index = False)
        if survey.verbose and not survey.emcee:
            print('   - final prioritized list saved to %s/observing_priorities.csv'%survey.path_save)
    survey.observed = observed.copy()
    return survey

        
def final_costs(survey, costs={}):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    tics = survey.final.tic.values.tolist()
    for i, tic in enumerate(survey.observed.tic.values.tolist()):
        frac = []
        costs[i] = {}
        costs[i]['priority'] = i+1
        costs[i]['toi'] = str(int(np.floor(survey.observed.loc[i, "toi"])))
        costs[i]['tic'] = str(int(tic))
        idx = tics.index(tic)
        nobs_goal = int(survey.final.loc[idx, "nobs_goal"])
        costs[i]['nobs_goal'] = str(nobs_goal)
        for science in survey.programs.index.values.tolist():
            cost = observing.cost_function(survey.final.loc[idx], survey.programs.loc[science]['method'])
            frac.append(cost*float(survey.final.loc[idx,'in_'+science]))
        if float(np.sum(frac)) != 0.:
            fractional = frac/np.sum(frac)
        else:
            fractional = np.zeros(len(frac))
        for f, science in zip(fractional, survey.programs.index.values.tolist()):
            costs[i][science] = round(((max(frac)/3600.)*f),3)
        costs[i]['charged_time'] = round((max(frac)/3600.),3)
        if nobs_goal == 60 or nobs_goal == 100:
            total_cost = observing.cost_function(survey.final.loc[idx], 'hires-nobs=%d-counts=ramp'%nobs_goal, include_archival=False)
        else:
            total_cost = observing.cost_function(survey.final.loc[idx], 'hires-nobs=%d-counts=60'%nobs_goal, include_archival=False)
        costs[i]['total_time'] = round(total_cost/3600.,3)
    costs = pd.DataFrame.from_dict(costs, orient = 'index')
    reorder = get_columns('costs', survey)
    df = pd.DataFrame(columns = reorder)
    for column in reorder:
        df[column] = costs[column]
    idx = len(df)
    total = df.sum(axis=0)
    df.loc[idx,'tic'] = 'Totals:'
    for science in survey.programs.index.values.tolist():
        df.loc[idx,science] = round(total[science],3)
    df.loc[idx,'charged_time'] = round(total['charged_time'],3)
    df.loc[idx,'total_time'] = round(total['total_time'],3)
    survey.charged_time = float(total['charged_time'])
    survey.total_time = float(total['total_time'])
    if survey.save:
        if survey.emcee:
            df.to_csv('%s/%d/total_costs.csv'%(survey.path_save,survey.n), index=False)
        else:
            df.to_csv('%s/total_costs.csv'%survey.path_save, index=False)
        if survey.verbose and not survey.emcee:
            print('   - final costs saved to %s/total_costs.csv'%survey.path_save)
    survey.costs = df.copy()
    return survey
        

def program_overlap(survey):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    columns = get_columns('overlap', survey)
    df = pd.DataFrame(columns = columns, index = survey.observed.index.values.tolist())
    for i in survey.observed.index.values.tolist():
        df_temp = survey.observed.loc[i]
        df.loc[i, 'tic'] = df_temp['tic']
        df.loc[i, 'toi'] = df_temp['toi']
        df.loc[i, 'priority'] = df_temp['overall_priority']
        for program in survey.programs.index.values.tolist():
            if program in df_temp['programs']:
                df.loc[i, 'in_'+program] = 'X'
            else:
                df.loc[i, 'in_'+program] = '-'
        df.loc[i, 'total_programs'] = len(df_temp['programs'])
    if survey.save:
        if survey.emcee:
            df.to_csv('%s/%d/program_overlap.csv'%(survey.path_save,survey.n))
        else:
            df.to_csv('%s/program_overlap.csv'%survey.path_save, index = False)
        if survey.verbose and not survey.emcee:
            print('   - spreadsheet containing program overlap saved to %s/program_overlap.csvs'%survey.path_save)
    survey.overlap = df.copy()
    return survey


def emcee_rankings(ranked):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    columns = get_columns('emcee', survey)
    selected = pd.DataFrame.from_dict(ranked, orient='index', columns=columns)
    for i in selected.index.values.tolist():
        prior = np.array(ranked[i]['priority'])
        check = len(prior)
        if check != selected.loc[i, 'n_select'] and survey.verbose:
            print('WARNING for TOI %d MC counts'%int(toi))
        selected.loc[i, 'mean'] = round(np.mean(prior),2)
        selected.loc[i, 'median'] = int(np.percentile(prior,50.))
    df = selected.sort_values(by = ['mean'], ascending = [True])
    for science in survey.programs.index.values.tolist():
        sample = Sample(df)
        relevant = sample.in_science(science)
        df[science] = relevant
    for i in df.index.values.tolist():
        total = 0
        for science in sciences:
            total += df.loc[i, science]
        df.loc[i, "other_programs"] = total
    df.to_csv('%s/TOIs_perfect_mc.csv'%survey.path_save)
    if survey.verbose:
        print('   - spreadsheet containing MC information saved to %s/TOIs_perfect_mc.csv'%survey.path_save)
    emcee_df = df.copy()
    return emcee_df
          
  
def get_stats(survey, note='', finish=False):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    if finish:
        df = pd.read_csv(survey.path_sample)
        df.query('finish == True', inplace=True)
        n_finish = len(df)
    df = survey.overlap.copy()
    time = datetime.datetime.now()
    note += time.strftime("%a %m/%d/%y %I:%M%p") + '\n'
    note += 'Seed no: %d\n'%survey.seeds[survey.n-1]
    note += 'Out of the %d total targets:\n'%len(df)
    for science in survey.programs.index.values.tolist():
        if science == 'TOA':
            note += '  - %s has %d targets\n'%(science, len(df))
        else:
            a = df['in_'+science].values.tolist()
            d = {x:a.count(x) for x in a}
            try:
                note += '  - %s has %d targets\n'%(science, d['X'])
            except KeyError:
                note += '  - %s has %d targets\n'%(science, 0)
    if survey.save:
        if survey.emcee:
            f = open('%s/%d/run_info.txt'%(survey.path_save,survey.n), "w")
        else:
            f = open('%s/run_info.txt'%survey.path_save, "w")
        f.write(note)
        f.close()
        if survey.verbose and not survey.emcee:
            print('   - txt file containing run information saved to %s/run_info.txt'%survey.path_save)
            print('')
            print(' --- process complete ---')
            print('')
            print(note)


def get_columns(type, survey):
    """
    Given a set of programs, selects a program randomly based on the proportional time remaining 
    for each program in a Survey. This is done by creating a cumulative distribution function (CDF) 
    using all programs "remaining_hours", normalizing by the total remaining time (Ttot), then draws 
    a random number from U~[0,1), which then maps back to the list of programs.

    Parameters
    ----------
    programs : dict
        program dictionary of Survey object (must have 'remaining_hours' as a key for each program)

    Returns
    -------
    program : str
        the selected program which comes directly from the input programs dict keys

    """
    if type == 'track':
        columns = ['program', 'program_pick', 'overall_priority', 'tic', 'toi'] + survey.sciences.name.values.tolist() + ['total_time']
    elif type == 'costs':
        columns = ['priority', 'tic', 'toi'] + survey.sciences.name.values.tolist() + ['nobs_goal', 'charged_time', 'total_time']
    elif type == 'overlap':
        columns = ['tic', 'toi', 'priority'] + ['in_%s'%program for program in survey.programs.index.values.tolist()] + ['total_programs']
    else:
        columns = []
    return columns


import os
import glob
import numpy as np
import pandas as pd


from sortasurvey import observing


class Sample:
    """
    Creates a filtered sample based on the provided survey program. The most common 
    use of the sample is during the target selection process, which calls the 
    get_highest_priority method at each iteration. All other methods are useful 
    for post-target selection filtering, shown in the provided Jupyter notebook example.

    Attributes
    ----------
    df : pandas.DataFrame
        vetted sample from the survey.Survey
    programs : pandas.DataFrame
        survey programs from the survey.Survey 

    Parameters
    ----------
    survey : Optional[survey.Survey]
        will save the science programs and vetted sample of a given survey to the
        Sample.programs and Sample.df attributes, respectively if not `None`.
    path_init : Optional[str]
        path to vetted sample (default = 'info/TOIs_perfect.csv')
    path_final : Optional[str]
        root path to results (default = `None`)
    program : str
        selected program of interest within a survey
    

    """

    def __init__(self, program, survey=None, path_init='info/TOIs_perfect.csv', path_final=None):
        self.path_init = path_init
        self.program = program
        if survey is not None:
            self.df = survey.candidates.copy()
            self.programs = survey.sciences.copy()
            self.get_vetted_science(program)


    def get_vetted_science(self, drop_dup=True):
        """
        Similar to the vetted sample method above, it loads the same sample which was (is)
        available for target selection but filters the sample to address a specific science 
        case's (or program) needs. 

        Attributes
        ----------
        program : str
            the program used to filter the selected Sample
        query : pandas.DataFrame
            the original vetted survey sample filtered on the specific program's selection criteria

        Parameters
        ----------
        drop_dup : Optional[bool]
            option to drop duplicate targets in the sample query, since the target will only be observed once. As a result, the default is `True`.

        """
        if not hasattr(self, 'df'):
            self.df = pd.read_csv(self.path_init)
        query = self.df.query(self.programs.loc[self.program,'filter'])
        if drop_dup:
            query = query.drop_duplicates(subset='tic')
        query = observing.get_actual_costs(self.program, self.programs, query)
        query = query.sort_values(by=self.programs.loc[self.program,'prioritize_by'], ascending=self.programs.loc[self.program,'ascending_by'])
        if self.programs.loc[self.program,'high_priority'] != []:
            top = pd.DataFrame(columns = query.columns.values.tolist())
            for toi in self.programs.loc[self.program,'high_priority']:
                new_filter = 'toi == %.2f'%toi
                new = self.df.query(new_filter)
                new = observing.get_actual_costs(self.program, self.programs, new)
                top = pd.concat([top,new])
            query = pd.concat([top,query])
        query.reset_index(drop=True, inplace=True)
        self.query = query.copy()
        

    def get_highest_priority(self):
        """
        Returns the highest priority target for a selected program that has not
        yet been selected by that program.

        Attributes
        ----------
        program : str
            the program used to filter the selected Sample
        query : pandas.DataFrame
            the original vetted survey sample filtered on the specific program's selection criteria

        Returns
        -------
        pick : pandas.DataFrame
            the selected's program's highest priority pick

        """
        for i in self.query.index.values.tolist():
            if not int(self.query.loc[i,'in_%s'%self.program]):
                pick = self.query.loc[i]
                break
        return pick


    def get_vetted_sample(self):
        """
        Loads the vetted sample that was available during the survey selection process.
        Using the final csv path, it selects the sample from the most recent output directory.

        Parameters
        ----------
        final_path : Optional[str]
            path of output csv file to read in

        Returns
        -------
        df : pandas.DataFrame
            the full vetted sample used for the target selection

        """
        list_of_files = glob.glob(self.final_path)
        latest_file = max(list_of_files, key=os.path.getctime)
        df = pd.read_csv(latest_file)
        return df


    def get_selected_sample(self, final_path=None):
        """
        Loads the final sample selected from the target prioritization process. If 
        no path is provided, it will automatically load in the most recent output.

        Parameters
        ----------
        final_path : Optional[str]
            path of output csv file to read in

        Returns
        -------
        df : pandas.DataFrame
            the sample selected by the algorithm for a given Survey 

        """
        if final_path is None:
            list_of_files = glob.glob('results/other/*/*/TOIs_perfect_final.csv')
            latest_file = max(list_of_files, key=os.path.getctime)
        else:
            latest_file = final_path
        df = pd.read_csv(latest_file)
        df.query("in_other_programs != 0", inplace=True)
        return df


    def get_science_sample(self, program, path_final=None):
        """
        Loads the final sample selected from the target prioritization process for a 
        specific science case or program from the Survey. If no path is provided, it
        will automatically load in the most recent output.

        Parameters
        ----------
        program : str
            specific Survey program to downselect the sample for
        path_final : Optional[str]
            path of output csv file to read in

        Returns
        -------
        df : pandas.DataFrame
            the sample selected by the algorithm for a specific Survey program

        """
        list_of_files = glob.glob(self.final_path)
        latest_file = max(list_of_files, key=os.path.getctime)
        df = pd.read_csv(latest_file)
        df.query("in_%s == 1"%program, inplace=True)
        return df


import os
import subprocess
import numpy as np
import pandas as pd
import time as clock
pd.set_option('mode.chained_assignment', None)


import sortasurvey
from sortasurvey import utils
from sortasurvey.survey import Survey
from sortasurvey.sample import Sample


def rank(args, stuck=0):

    # init Survey class
    survey = Survey(args)
    ti = clock.time()
    # Monte-Carlo simulations of sampler (args.iter=1 by default)
    for n in range(1,args.iter+1):
        survey.n = n
        survey.reset_track()
        # Begin selection process and continue while the total remaining 
        # time allocation is greater than zero
        while np.sum(survey.sciences.remaining_hours.values.tolist()) > 0.:
            # Select program
            program = utils.pick_program(survey.sciences)
            # Create an instance åof the Sample class w/ the updated vetted sample
            sample = Sample(program, survey=survey)
            # Only continue if the selected program has targets left
            if not survey.sciences.loc[program,'n_targets_left']:
                continue
            if program in survey.special:
                survey.get_2D_filter(program)
            # pick highest priority target not yet selected
            pick = sample.get_highest_priority()
            # what is the cost of the selected target
            cost = float((pick.actual_cost))/3600.
            # if the program cannot afford the target, it is "stuck"
            if cost > survey.sciences.loc[program,'remaining_hours']:
                if program in survey.special:
                    survey.special[program]['stuck'] += 1
                stuck += 1
            else:
                stuck = 0
                # update records with the program pick
                survey.update(pick, program)
            if stuck >= len(survey.sciences):
                break
        if survey.emcee:
            survey.df = survey.candidates.copy()
            utils.make_data_products(survey)

    tf = clock.time()
    survey.ranking_time = float(tf-ti)
    survey.df = survey.candidates.copy()
    utils.make_data_products(survey)


def setup(args, note='', source='https://raw.githubusercontent.com/ashleychontos/sort-a-survey/main/info'):
    """
    Running this after installation will create the appropriate directories in the current working
    directory as well as download example files to test your installation.

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    note : Optional[str]
        suppressed verbose output
    source : Optional[str]
        source directory to download example files to test your installation

    """

    print('\n\nDownloading relevant data from source directory:\n')
    note+='NB:\n'
    # create info directory
    if not os.path.exists(args.inpdir):
        os.mkdir(args.inpdir)
        note+=' - created input file directory: %s \n'%args.inpdir

    # get example TKS input files
    for file in ['/TOIs_perfect.csv', '/high_priority.csv', '/no_no.csv', '/survey_info.csv']:
        infile='%s%s'%(source,file)
        outfile='%s%s'%(args.inpdir,file)
        subprocess.call(['curl %s > %s'%(infile, outfile)], shell=True)

    # create results directory
    if not os.path.exists(args.outdir):
        os.mkdir(args.outdir)
    note+=' - results will be saved to %s \n\n'%args.outdir
    
    if args.verbose:
        print(note)


import os

from .observing import *
from .pipeline import *
from .sample import *
from .survey import *
from .utils import *

__all__ = ['cli', 'observing', 'pipeline', 'sample', 'survey', 'utils']

__version__ = '0.0.7'


_ROOT = os.path.abspath(os.getcwd())
INPDIR = os.path.join(_ROOT, 'info')
OUTDIR = os.path.join(_ROOT, 'results')